#
#  Epour - A bittorrent client using EFL and libtorrent
#
#  Copyright 2012-2013 Kai Huuhko <kai.huuhko@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import os
import cgi
import logging
from datetime import timedelta

import libtorrent as lt

try:
    from evas import EVAS_ASPECT_CONTROL_VERTICAL, Rectangle
    from ecore import Timer
    import elementary as elm
    from elementary import Genlist, GenlistItemClass, StandardWindow, Icon, \
        Box, Label, Button, ELM_GENLIST_ITEM_FIELD_TEXT, \
        ELM_GENLIST_ITEM_FIELD_CONTENT, InnerWindow, Frame, \
        Fileselector, Entry, Panel, ELM_PANEL_ORIENT_BOTTOM, \
        ELM_OBJECT_SELECT_MODE_NONE, Table, Separator, Menu, Configuration
except ImportError:
    from efl.evas import EVAS_ASPECT_CONTROL_VERTICAL, Rectangle
    from efl.ecore import Timer
    from efl import elementary as elm
    from efl.elementary.genlist import Genlist, GenlistItemClass, \
        ELM_GENLIST_ITEM_FIELD_TEXT, ELM_GENLIST_ITEM_FIELD_CONTENT, \
        ELM_OBJECT_SELECT_MODE_NONE
    from efl.elementary.window import StandardWindow
    from efl.elementary.icon import Icon
    from efl.elementary.box import Box
    from efl.elementary.label import Label
    from efl.elementary.button import Button
    from efl.elementary.innerwindow import InnerWindow
    from efl.elementary.frame import Frame
    from efl.elementary.fileselector import Fileselector
    from efl.elementary.entry import Entry
    from efl.elementary.panel import Panel, ELM_PANEL_ORIENT_BOTTOM
    from efl.elementary.table import Table
    from efl.elementary.separator import Separator
    from efl.elementary.menu import Menu
    from efl.elementary.configuration import Configuration

from TorrentInfo import TorrentInfo
from Preferences import Preferences
from Notify import ConfirmExit, Error, Information

from intrepr import intrepr

class MainInterface(object):
    def __init__(self, parent, session):
        self.parent = parent
        self.session = session

        elm_conf = Configuration()
        scale = elm_conf.scale

        self.log = logging.getLogger("epour.gui")

        self.torrentitems = {}

        win = self.win = StandardWindow("epour", "Epour")
        win.callback_delete_request_add(lambda x: elm.exit())
        win.screen_constrain = True
        win.size = 480 * scale, 400 * scale

        self.tlist = tlist = Genlist(win)
        tlist.select_mode_set(ELM_OBJECT_SELECT_MODE_NONE)
        tlist.callback_activated_add(self.item_activated_cb)
        tlist.homogeneous_set(True)
        tlist.size_hint_weight_set(1.0, 1.0)
        tlist.size_hint_align_set(-1.0, -1.0)
        tlist.show()

        mbox = Box(win)
        mbox.size_hint_weight = 1.0, 1.0
        win.resize_object_add(mbox)
        mbox.show()

        mbox.pack_end(tlist)

        pad = Rectangle(win.evas)
        pad.size_hint_weight = 1.0, 1.0

        p = Panel(win)
        #p.tooltip_text_set("Session")
        p.color = 200,200,200,200
        p.size_hint_weight = 1.0, 1.0
        p.size_hint_align = -1.0, -1.0
        p.orient = ELM_PANEL_ORIENT_BOTTOM
        p.content = SessionStatus(win, session)
        p.hidden = True
        p.show()

        topbox = Box(win)
        topbox.horizontal = True
        topbox.size_hint_weight = 1.0, 1.0
        win.resize_object_add(topbox)

        topbox.pack_end(pad)
        topbox.pack_end(p)
        topbox.stack_above(mbox)
        topbox.show()


        for a in lt.torrent_paused_alert, lt.torrent_resumed_alert:
            session.alert_manager.callback_add(a, self.update_icon)
        session.alert_manager.callback_add(lt.state_changed_alert, self.state_changed_cb)
        session.alert_manager.callback_add(lt.torrent_finished_alert, self.torrent_finished_cb)

    def state_changed_cb(self, a):
        h = a.handle
        ihash = str(h.info_hash())
        if not h.is_valid():
            self.remove_torrent_item(h)

        # FIXME: Find out if this can be used instead of the missing
        #        torrent_added alert in libtorrent 0.15.x
        #elif not self.torrentitems.has_key(ihash):
            #self.add_torrent_item(h)

        self.update_icon(a)

    def run(self):
        self.win.show()

        self.timer = Timer(1.0, self.update)
        elm.run()
        self.quit()

    def update(self):
        for v in self.tlist.realized_items_get():
            v.fields_update("*", ELM_GENLIST_ITEM_FIELD_TEXT)
        return True

    def update_icon(self, a):
        h = a.handle
        if not h.is_valid(): return
        ihash = str(h.info_hash())
        if not self.torrentitems.has_key(ihash): return
        self.torrentitems[ihash].fields_update(
            "elm.swallow.icon", ELM_GENLIST_ITEM_FIELD_CONTENT
        )

    def add_torrent_item(self, h):
        ihash = str(h.info_hash())

        itc = TorrentClass("double_label")
        item = self.tlist.item_append(itc, h)
        self.torrentitems[ihash] = item

    def remove_torrent_item(self, h):
        ihash = str(h.info_hash())
        if self.torrentitems.has_key(ihash):
            self.torrentitems[ihash].delete()

    def item_activated_cb(self, gl, item):
        h = item.data
        menu = Menu(self.win)

        menu.item_add(
            None,
            "Resume" if h.is_paused() else "Pause",
            None,
            self.resume_torrent_cb if h.is_paused() else self.pause_torrent_cb,
            h
        )
        q = menu.item_add(None, "Queue", None, None)
        menu.item_add(q, "Up", None, lambda x, y: h.queue_position_up())
        menu.item_add(q, "Down", None, lambda x, y: h.queue_position_down())
        menu.item_add(q, "Top", None, lambda x, y: h.queue_position_top())
        menu.item_add(q, "Bottom", None, lambda x, y: h.queue_position_bottom())
        rem = menu.item_add(None, "Remove torrent", None,
            self.remove_torrent_cb, item, h, False)
        menu.item_add(rem, "and data files", None,
            self.remove_torrent_cb, item, h, True)
        menu.item_add(None, "Force re-check", None,
            self.force_recheck, h)
        menu.item_separator_add(None)
        menu.item_add(None, "Torrent preferences", None,
            self.torrent_preferences_cb, h)

        menu.move(*self.win.evas.pointer_canvas_xy_get())
        menu.show()

    def resume_torrent_cb(self, menu, item, h):
        h.resume()
        h.auto_managed(True)

    def pause_torrent_cb(self, menu, item, h):
        h.auto_managed(False)
        h.pause()

    def force_recheck(self, menu, item, h):
        h.force_recheck()

    def remove_torrent_cb(self, menu, item, glitem, h, with_data=False):
        menu.close()
        ihash = self.parent.session.remove_torrent(h, with_data)
        glitem.delete()
        del self.torrentitems[ihash]
        del glitem

    def torrent_preferences_cb(self, menu, item, h):
        self.i = TorrentInfo(self, h)

    def show_error(self, title, text):
        Error(self.win, title, text)

    def torrent_finished_cb(self, a):
        msg = "Torrent {} has finished downloading.".format(
            cgi.escape(a.handle.name())
        )
        self.log.info(msg)

        Information(self.win, msg)

    def quit(self, *args):
        if self.session.conf.getboolean("Settings", "confirmations"):
            ConfirmExit(self.win, self.shutdown())
        else:
            self.shutdown()

    def shutdown(self):
        elm.shutdown()
        self.parent.quit()

class SessionStatus(Table):

    log = logging.getLogger("epour.gui.session")

    def __init__(self, parent, session):
        Table.__init__(self, parent)
        self.session = session

        s = session.status()

        self.padding = 5, 5

        ses_pause_ic = self.ses_pause_ic = Icon(parent)
        ses_pause_ic.size_hint_align = -1.0, -1.0
        try:
            ses_pause_ic.standard = "player_play" if session.is_paused() else "player_play"
        except RuntimeError:
            self.log.debug("Setting session ic failed")
        self.pack(ses_pause_ic, 1, 0, 1, 1)
        ses_pause_ic.show()

        title_l = Label(parent)
        title_l.text = "<b>Session</b>"
        self.pack(title_l, 0, 0, 1, 1)
        title_l.show()

        d_ic = Icon(parent)
        try:
            d_ic.standard = "down"
        except RuntimeError:
            self.log.debug("Setting d_ic failed")
        d_ic.size_hint_align = -1.0, -1.0
        self.pack(d_ic, 0, 2, 1, 1)
        d_ic.show()

        d_l = self.d_l = Label(parent)
        d_l.text = "{}/s".format(intrepr(s.payload_download_rate))
        self.pack(d_l, 1, 2, 1, 1)
        d_l.show()

        u_ic = Icon(self)
        try:
            u_ic.standard = "up"
        except RuntimeError:
            self.log.debug("Setting u_ic failed")
        u_ic.size_hint_align = -1.0, -1.0
        self.pack(u_ic, 0, 3, 1, 1)
        u_ic.show()

        u_l = self.u_l = Label(parent)
        u_l.text = "{}/s".format(intrepr(s.payload_upload_rate))
        self.pack(u_l, 1, 3, 1, 1)
        u_l.show()

        peer_t = Label(parent)
        peer_t.text = "Peers"
        self.pack(peer_t, 0, 4, 1, 1)
        peer_t.show()

        peer_l = self.peer_l = Label(parent)
        peer_l.text = str(s.num_peers)
        self.pack(peer_l, 1, 4, 1, 1)
        peer_l.show()


        addbtn = Button(parent)
        addbtn.text_set("Add torrent")
        addbtn.callback_clicked_add(self.select_torrent)
        addbtn.show()

        togglebtn = self.togglebtn = Button(parent)
        togglebtn.text = "Resume" if \
            self.session.is_paused() else "Pause"
        togglebtn.callback_clicked_add(self.toggle_paused_cb)
        togglebtn.show()

        prefbtn = Button(parent)
        prefbtn.text = "Preferences"
        prefbtn.callback_clicked_add(self.prefs_cb)
        prefbtn.show()

        btnbox = Box(parent)
        btnbox.horizontal = True
        btnbox.size_hint_align = -1.0, 0.0
        btnbox.pack_end(addbtn)
        btnbox.pack_end(togglebtn)
        btnbox.pack_end(prefbtn)
        btnbox.show()

        self.pack(btnbox, 0, 10, 2, 1)


        self.show()

        self.update_timer = Timer(1.0, self.update)

    def update(self):
        s = self.session.status()
        self.d_l.text = "{}/s".format(intrepr(s.payload_download_rate))
        self.u_l.text = "{}/s".format(intrepr(s.payload_upload_rate))
        self.peer_l.text = str(s.num_peers)
        return True

    def toggle_paused_cb(self, btn):
        session = self.session
        if session.is_paused():
            session.resume()
            btn.text = "Pause"
            try:
                self.ses_pause_ic.standard = "player_play"
            except RuntimeError:
                self.log.debug("")
        else:
            session.pause()
            btn.text = "Resume"
            try:
                self.ses_pause_ic.standard = "player_pause"
            except RuntimeError:
                self.log.debug("")

    def select_torrent(self, btn):
        sel = Fileselector(self.top_widget)
        sel.expandable = False
        sel.path_set(os.path.expanduser("~"))
        sel.size_hint_weight_set(1.0, 1.0)
        sel.size_hint_align_set(-1.0, -1.0)
        sel.show()
        sf = Frame(self.top_widget)
        sf.size_hint_weight_set(1.0, 1.0)
        sf.size_hint_align_set(-1.0, -1.0)
        sf.text = "Select torrent file"
        sf.content = sel
        sf.show()

        magnet = Entry(self.top_widget)
        magnet.single_line = True
        magnet.scrollable = True
        magnet.show()
        mf = Frame(self.top_widget)
        mf.size_hint_weight_set(1.0, 0.0)
        mf.size_hint_align_set(-1.0, 0.0)
        mf.text = "Or enter magnet URI here"
        mf.content = magnet
        mf.show()
        mbtn = Button(self.top_widget)
        mbtn.text = "Done"
        mbtn.show()
        mbox = Box(self.top_widget)
        mbox.size_hint_weight_set(1.0, 0.0)
        mbox.size_hint_align_set(-1.0, 0.0)
        mbox.horizontal = True
        mbox.pack_end(mf)
        mbox.pack_end(mbtn)
        mbox.show()

        box = Box(self.top_widget)
        box.size_hint_weight = (1.0, 1.0)
        box.size_hint_align = (-1.0, -1.0)
        box.pack_end(sf)
        box.pack_end(mbox)
        box.show()

        self.inwin = inwin = InnerWindow(self.top_widget)
        inwin.content = box
        sel.callback_done_add(self.add_torrent_cb)
        sel.callback_done_add(lambda x, y: inwin.delete())
        mbtn.callback_clicked_add(self.add_magnet_uri_cb, magnet)
        mbtn.callback_clicked_add(lambda x: inwin.delete())
        inwin.activate()

    def prefs_cb(self, btn):
        self.prefs = Preferences(self, self.session)

    def add_torrent_cb(self, filesel, t):
        if not t:
            return

        self.session.add_torrent(t)

    def add_magnet_uri_cb(self, btn, magnet):
        m = magnet.text
        if m:
            self.add_torrent_cb(None, m)


class TorrentClass(GenlistItemClass):

    state_str = ['Queued', 'Checking', 'Downloading metadata', \
        'Downloading', 'Finished', 'Seeding', 'Allocating', \
        'Checking resume data']

    log = logging.getLogger("epour.gui.torrent_list")

    def text_get(self, obj, part, item_data):
        h = item_data
        name = h.name()

        if part == "elm.text":
            return '%s' % (
                name
            )
        elif part == "elm.text.sub":
            s = h.status()

            return '{:.0%} complete, ETA: {} (Down: {}/s Up: {}/s Peers: {} Queue: {})'.format(
                s.progress,
                timedelta(seconds=self.get_eta(h)),
                intrepr(s.download_payload_rate, precision=0),
                intrepr(s.upload_payload_rate, precision=0),
                s.num_peers,
                h.queue_position(),
            )

    def content_get(self, obj, part, item_data):
        if part == "elm.swallow.icon":
            h = item_data
            s = h.status()
            ic = Icon(obj)
            if h.is_paused():
                try:
                    ic.standard = "player_pause"
                except RuntimeError:
                    self.log.debug("Setting torrent ic failed")
            elif h.is_seed():
                try:
                    ic.standard = "up"
                except RuntimeError:
                    self.log.debug("Setting torrent ic failed")
            else:
                try:
                    ic.standard = "down"
                except RuntimeError:
                    self.log.debug("Setting torrent ic failed")
            ic.tooltip_text_set(self.state_str[s.state])
            ic.size_hint_aspect_set(EVAS_ASPECT_CONTROL_VERTICAL, 1, 1)
            return ic

    def get_eta(self, h):
        s = h.status()
        if False: #self.is_finished and self.options["stop_at_ratio"]:
            # We're a seed, so calculate the time to the 'stop_share_ratio'
            if not s.upload_payload_rate:
                return 0
            stop_ratio = self.options["stop_ratio"]
            return ((s.all_time_download * stop_ratio) - s.all_time_upload) / s.upload_payload_rate

        left = s.total_wanted - s.total_wanted_done

        if left <= 0 or s.download_payload_rate == 0:
            return 0

        try:
            eta = left / s.download_payload_rate
        except ZeroDivisionError:
            eta = 0

        return eta
