#
#  Epour - A bittorrent client using EFL and libtorrent
#
#  Copyright 2012-2013 Kai Huuhko <kai.huuhko@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#

import os
import cgi
import logging
log = logging.getLogger("epour")

import libtorrent as lt

try:
    from elementary import Icon, Box, Label, Button, \
        InnerWindow, Frame, Entry, Check, Spinner, Hoversel, \
        FileselectorButton, Scroller, ELM_SCROLLER_POLICY_OFF, \
        ELM_SCROLLER_POLICY_AUTO, Separator, Slider, Actionslider, \
        ELM_ACTIONSLIDER_LEFT, ELM_ACTIONSLIDER_CENTER, \
        ELM_ACTIONSLIDER_RIGHT, ELM_ACTIONSLIDER_ALL, \
        Naviframe, Table, Configuration
    from evas import Rectangle
    from ecore import Timer
except ImportError:
    from efl.elementary.icon import Icon
    from efl.elementary.box import Box
    from efl.elementary.label import Label
    from efl.elementary.button import Button
    from efl.elementary.innerwindow import InnerWindow
    from efl.elementary.frame import Frame
    from efl.elementary.entry import Entry
    from efl.elementary.check import Check
    from efl.elementary.spinner import Spinner
    from efl.elementary.hoversel import Hoversel
    from efl.elementary.fileselector_button import FileselectorButton
    from efl.elementary.scroller import Scroller, ELM_SCROLLER_POLICY_OFF, \
        ELM_SCROLLER_POLICY_AUTO
    from efl.elementary.separator import Separator
    from efl.elementary.slider import Slider
    from efl.elementary.actionslider import Actionslider, \
        ELM_ACTIONSLIDER_LEFT, ELM_ACTIONSLIDER_CENTER, \
        ELM_ACTIONSLIDER_RIGHT, ELM_ACTIONSLIDER_ALL
    from efl.elementary.naviframe import Naviframe
    from efl.elementary.table import Table
    from efl.elementary.configuration import Configuration
    from efl.evas import Rectangle
    from efl.ecore import Timer

import Notify

class Preferences(InnerWindow):
    def __init__(self, parent, session):
        self.session = session
        conf = session.conf
        InnerWindow.__init__(self, parent.top_widget)

        nf = Naviframe(self)
        nf.size_hint_weight = 1.0, 1.0

        mbox = Box(self)
        mbox.size_hint_weight = 1.0, 1.0
        mbox.size_hint_align = -1.0, -1.0

        lbl = Label(self)
        lbl.text = "Click on a title to open/close configuration entry."

        limits = Limits(self, session)
        ports = ListenPorts(self, session)
        pf = ProxyConfig(self)
        pe = EncryptionSettings(self, session)
        dlsel = DataStorageSelector(self, conf)

        pad = Rectangle(self.evas)
        pad.color = 0, 0, 0, 0
        pad.size_hint_min = 0, 10

        sep1 = Separator(self)
        sep1.horizontal = True

        chk1 = Check(self)
        chk1.size_hint_align = 0.0, 0.0
        chk1.text = "Delete original .torrent file when added"
        chk1.state = conf.getboolean("Settings", "delete_original")
        chk1.callback_changed_add(lambda x: conf.set("Settings",
            "delete_original", str(bool(chk1.state))))

        chk2 = Check(self)
        chk2.size_hint_align = 0.0, 0.0
        chk2.text = "Ask for confirmation on exit"
        chk2.state = conf.getboolean("Settings", "confirmations")
        chk2.callback_changed_add(lambda x: conf.set("Settings",
            "confirmations", str(bool(chk2.state))))

        sep2 = Separator(self)
        sep2.horizontal = True

        ses_btn = Button(self)
        ses_btn.text = "Session settings"
        ses_btn.callback_clicked_add(self.show_ses_set, nf, session)

        xbtn = Button(self)
        xbtn.text_set("Close")
        xbtn.callback_clicked_add(lambda x: self.delete())

        for w in lbl, ports, limits, pf, dlsel, pe, pad, \
            sep1, chk1, chk2, sep2, ses_btn, xbtn:
            w.show()
            mbox.pack_end(w)

        scr = Scroller(self)
        scr.policy = ELM_SCROLLER_POLICY_OFF, ELM_SCROLLER_POLICY_AUTO
        scr.content = mbox

        nf.item_simple_push(scr)
        nf.show()

        self.content = nf
        self.activate()

    def show_ses_set(self, btn, nf, session):
        ses_set = SessionSettings(self, session, nf)
        nf.item_simple_push(ses_set)

class DataStorageSelector(Frame):
    def __init__(self, parent, conf):
        Frame.__init__(self, parent)

        self.size_hint_align = -1.0, 0.0
        self.size_hint_weight = 1.0, 0.0
        self.text = "Data storage"
        self.collapse = True
        self.autocollapse = True

        self.conf = conf

        b = Box(parent)

        lbl = self.path_lbl = Label(parent)
        lbl.text = conf.get("Settings", "storage_path")

        self.dlsel = dlsel = FileselectorButton(self)
        dlsel.size_hint_align = -1.0, 0.0
        dlsel.inwin_mode = True
        dlsel.folder_only = True
        dlsel.expandable = False
        dlsel.text = "Change path"
        dlsel.path = conf.get("Settings", "storage_path")
        dlsel.callback_file_chosen_add(self.save_dlpath)

        for w in lbl, dlsel:
            w.show()
            b.pack_end(w)

        b.show()
        self.content = b

    def save_dlpath(self, fs, path):
        if not path:
            return

        if not os.path.exists(self.dlsel.path):
            p = Notify.Error(self, "Invalid storage path",
                "You have selected an invalid data storage path for torrents.")
            return

        self.path_lbl.text = path
        self.conf.set("Settings", "storage_path", self.dlsel.path)

class ListenPorts(Frame):
    def __init__(self, parent, session):
        Frame.__init__(self, parent)

        self.session = session

        self.size_hint_align = -1.0, -1.0
        self.collapse = True
        self.autocollapse = True
        self.text = "Listen port (range)"

        port = session.listen_port()

        b = Box(parent)

        lp = self.lp = RangeSpinners(
            parent,
            low = session.conf.getint("Settings", "listen_low"),
            high = session.conf.getint("Settings", "listen_high"),
            minim = 0, maxim = 65535)
        lp.show()
        b.pack_end(lp)

        save = Button(parent)
        save.text = "Apply"
        save.callback_clicked_add(self.save_cb)
        save.show()
        b.pack_end(save)

        b.show()

        self.content = b

    def save_cb(self, btn):
        low = int(self.lp.listenlow.value)
        high = int(self.lp.listenhigh.value)
        self.session.listen_on(low, high)
        self.session.conf.set("Settings", "listen_low", str(low))
        self.session.conf.set("Settings", "listen_high", str(high))

class ProxyConfig(Frame):
    def __init__(self, parent):
        Frame.__init__(self, parent)

        session = self.session = parent.session

        self.collapse = True
        self.autocollapse = True
        self.size_hint_weight = (1.0, 0.0)
        self.size_hint_align = (-1.0, 0.0)
        self.text = "Proxy configuration"

        b = Box(parent)

        proxies = [
            ["Proxy for torrent peer connections",
                session.peer_proxy, session.set_peer_proxy],
            ["Proxy for torrent web seed connections",
                session.web_seed_proxy, session.set_web_seed_proxy],
            ["Proxy for tracker connections",
                session.tracker_proxy, session.set_tracker_proxy],
            ["Proxy for DHT connections",
                session.dht_proxy, session.set_dht_proxy],
        ]

        for title, rfunc, wfunc in proxies:
            pg = ProxyGroup(parent, title, rfunc, wfunc)
            pg.show()
            b.pack_end(pg)

        b.show()
        self.content = b

class ProxyGroup(Frame):

    proxy_types = {
        lt.proxy_type.none.name: lt.proxy_type.none,
        lt.proxy_type.socks4.name: lt.proxy_type.socks4,
        lt.proxy_type.socks5.name: lt.proxy_type.socks5,
        lt.proxy_type.socks5_pw.name: lt.proxy_type.socks5_pw,
        lt.proxy_type.http.name: lt.proxy_type.http,
        lt.proxy_type.http_pw.name: lt.proxy_type.http_pw,
    }

    def __init__(self, parent, title, rfunc, wfunc):
        Frame.__init__(self, parent)

        self.collapse = True
        self.autocollapse = True
        self.size_hint_weight = 1.0, 0.0
        self.size_hint_align = -1.0, 0.0
        self.text = title

        b = Box(parent)

        f = Frame(parent)
        f.size_hint_align = -1.0, 0.0
        f.text = "Proxy type"
        ptype = Hoversel(parent)
        ptype.text = rfunc().type.name
        for n in self.proxy_types.iterkeys():
            ptype.item_add(n, callback=lambda x, y, z=n: ptype.text_set(z))
        f.content = ptype
        f.show()
        b.pack_end(f)

        f = Frame(parent)
        f.size_hint_align = -1.0, 0.0
        f.text = "Hostname"
        phost = Entry(parent)
        phost.single_line = True
        phost.entry = rfunc().hostname
        phost.show()
        f.content = phost
        f.show()
        b.pack_end(f)

        f = Frame(parent)
        f.size_hint_align = -1.0, 0.0
        f.text = "Port"
        pport = Spinner(parent)
        pport.min_max = 0, 65535
        pport.value = rfunc().port
        pport.show()
        f.content = pport
        f.show()
        b.pack_end(f)

        f = Frame(parent)
        f.size_hint_align = -1.0, 0.0
        f.text = "Username"
        puser = Entry(parent)
        puser.single_line = True
        puser.entry = rfunc().username
        puser.show()
        f.content = puser
        f.show()
        b.pack_end(f)

        f = Frame(parent)
        f.size_hint_align = -1.0, 0.0
        f.text = "Password"
        ppass = Entry(parent)
        ppass.single_line = True
        ppass.password = True
        ppass.entry = rfunc().password
        ppass.show()
        f.content = ppass
        f.show()
        b.pack_end(f)

        entries = [ptype, phost, pport, puser, ppass]

        save = Button(parent)
        save.callback_clicked_add(self.save_conf, wfunc, entries)
        save.text = "Apply"
        save.show()
        b.pack_end(save)

        self.content = b

    def save_conf(self, btn, wfunc, entries):
        ptype, phost, pport, puser, ppass = entries
        p = lt.proxy_settings()

        p.hostname = phost.entry.encode("utf-8")
        p.port = int(pport.value)
        p.username = puser.entry.encode("utf-8")
        p.password = ppass.entry.encode("utf-8")
        p.type = self.proxy_types[ptype.text]

        wfunc(p)

class EncryptionSettings(Frame):
    def __init__(self, parent, session):
        self.session = session

        Frame.__init__(self, parent)
        self.collapse = True
        self.autocollapse = True
        self.size_hint_align = -1.0, 0.0
        self.text = "Encryption settings"

        pes = self.pes = session.get_pe_settings()

        b = Box(parent)

        enc_values = lt.enc_policy.disabled, lt.enc_policy.enabled, lt.enc_policy.forced
        enc_levels = lt.enc_level.plaintext, lt.enc_level.rc4, lt.enc_level.both

        inc = self.inc = ActSWithLabel(parent,
            "Incoming encryption", enc_values, pes.in_enc_policy)
        b.pack_end(inc)
        inc.show()

        out = self.out = ActSWithLabel(parent,
            "Outgoing encryption", enc_values, pes.out_enc_policy)
        b.pack_end(out)
        out.show()

        lvl = self.lvl = ActSWithLabel(parent,
            "Allowed encryption level", enc_levels, pes.allowed_enc_level)
        b.pack_end(lvl)
        lvl.show()

        prf = self.prf = Check(parent)
        prf.style = "toggle"
        prf.text = "Prefer RC4 ecryption"
        prf.state = pes.prefer_rc4
        b.pack_end(prf)
        prf.show()

        a_btn = Button(parent)
        a_btn.text = "Apply"
        a_btn.callback_clicked_add(self.apply)
        b.pack_end(a_btn)
        a_btn.show()

        b.show()
        self.content = b

    def apply(self, btn):
        #TODO: Use callbacks to set these?
        self.pes.in_enc_policy = self.inc.get_value()
        self.pes.out_enc_policy = self.out.get_value()

        #FIXME: Find out why this isn't saved to the session.
        self.pes.allowed_enc_level = self.lvl.get_value()
        self.pes.prefer_rc4 = self.prf.state

        self.session.set_pe_settings(self.pes)

class ActSWithLabel(Box):
    def __init__(self, parent, label_text, values, initial_value):
        Box.__init__(self, parent)

        self.vd = {
            ELM_ACTIONSLIDER_LEFT: values[0],
            ELM_ACTIONSLIDER_CENTER: values[1],
            ELM_ACTIONSLIDER_RIGHT: values[2],
        }

        self.horizontal = True
        self.size_hint_align = -1.0, 0.0
        self.size_hint_weight = 1.0, 0.0

        l = Label(parent)
        l.text = label_text
        l.show()
        w = self.w = Actionslider(parent)
        w.magnet_pos = ELM_ACTIONSLIDER_ALL
        w.size_hint_align = -1.0, 0.0
        w.size_hint_weight = 1.0, 0.0
        w.show()

        parts = "left", "center", "right"

        for i, v in enumerate(values):
            w.part_text_set(parts[i], str(v))
        w.indicator_pos = values.index(initial_value) + 1

        self.pack_end(l)
        self.pack_end(w)

    def get_value(self):
        return self.vd[self.w.indicator_pos]

class SessionSettings(Scroller):
    def __init__(self, parent, session, nf):
        Scroller.__init__(self, parent)

        # TODO: Construct and populate this with an Idler

        self.canvas = parent
        self.session = session

        widgets = {}

        elm_conf = Configuration()

        s = session.settings()

        t = Table(parent)
        t.padding = 5, 5
        t.size_hint_weight = 1.0, 1.0
        t.size_hint_align = -1.0, -1.0

        i = 0

        for k in dir(s):
            if k.startswith("__"): continue
            try:
                a = getattr(s, k)
                if isinstance(a, lt.disk_cache_algo_t):
                    w = Check(parent)
                    w.size_hint_align = 1.0, 0.0
                    w.style = "toggle"
                    w.part_text_set("off", "lru")
                    w.part_text_set("on", "largest_continuous")
                    w.state = a
                elif isinstance(a, bool):
                    w = Check(parent)
                    w.size_hint_align = 1.0, 0.0
                    w.style = "toggle"
                    w.state = a
                elif isinstance(a, int):
                    w = Spinner(parent)
                    w.value = a
                elif isinstance(a, float):
                    w = Slider(parent)
                    w.size_hint_align = -1.0, 0.0
                    w.size_hint_weight = 1.0, 0.0
                    w.unit_format = "%1.2f"
                    if k.startswith("peer_turnover"):
                        w.min_max = 0.0, 1.0
                    else:
                        w.min_max = 0.0, 20.0
                    w.value = a
                elif k == "peer_tos":
                    # XXX: This is an int pair in libtorrent,
                    #      which doesn't have a python equivalent.
                    continue
                elif k == "user_agent":
                    w = Entry(parent)
                    w.size_hint_align = 1.0, 0.0
                    w.size_hint_weight = 1.0, 0.0
                    w.single_line = True
                    w.editable = False
                    w.entry = cgi.escape(a)
                else:
                    w = Entry(parent)
                    w.size_hint_align = -1.0, 0.0
                    w.size_hint_weight = 1.0, 0.0
                    w.single_line = True
                    w.entry = cgi.escape(a)
                l = Label(parent)
                l.text = k.replace("_", " ").capitalize()
                l.size_hint_align = 0.0, 0.0
                l.size_hint_weight = 1.0, 0.0
                l.show()
                t.pack(l, 0, i, 1, 1)
                w.size_hint_min = elm_conf.scale * 150, elm_conf.scale * 25
                w.size_hint_align = -1.0, 0.5
                w.size_hint_weight = 1.0, 0.0
                w.show()
                t.pack(w, 1, i, 1, 1)
                widgets[k] = w
                i += 1
            except TypeError:
                pass #print("Error {}".format(k))


        save_btn = Button(parent)
        save_btn.text = "Apply session settings"
        save_btn.callback_clicked_add(self.apply_settings, widgets, session)
        save_btn.show()
        t.pack(save_btn, 0, i, 2, 1)
        i += 1

        xbtn = Button(parent)
        xbtn.text = "Close"
        xbtn.callback_clicked_add(lambda x: nf.item_pop())
        xbtn.show()
        t.pack(xbtn, 0, i, 2, 1)
        i += 1

        self.content = t
        t.show()

    def apply_settings(self, btn, widgets, session):
        s = lt.session_settings()

        for k, w in widgets.iteritems():

            if k == "disk_cache_algorithm":
                v = lt.disk_cache_algo_t(w.state)
            elif isinstance(w, Spinner):
                v = int(w.value)
            elif isinstance(w, Slider):
                v = w.value
            elif isinstance(w, Entry):
                v = w.entry.encode("utf-8")
            elif isinstance(w, Check):
                v = bool(w.state)
            else:
                v = None

            setattr(s, k, v)

        session.set_settings(s)
        Notify.Information(self.canvas, "Session settings saved.")

class UnitSpinner(Box):
    def __init__(self, parent, base, units):
        self.base = base # the divisor/multiplier for units
        self.units = units # a list of strings with the base unit description at index 0

        Box.__init__(self, parent)
        self.horizontal = True

        s = self.spinner = Spinner(parent)
        s.show()
        self.pack_end(s)

        hs = self.hoversel = Hoversel(parent)
        for u in units:
            hs.item_add(u, None, 0, lambda x=hs, y=None, u=u: x.text_set(u))
        hs.show()
        self.pack_end(hs)

    def get_value(self):
        return self.spinner.value * ( self.base ** self.units.index(self.hoversel.text) )

    def set_value(self, v):
        i = 0

        while v // self.base > 0:
            i += 1
            v = float(v) / float(self.base)

        if i > len(self.units):
            i = len(self.units) - 1

        self.spinner.value = v
        self.hoversel.text = self.units[i]

class RangeSpinners(Box):
    def __init__(self, parent, low, high, minim, maxim):

        Box.__init__(self, parent)
        self.horizontal = True

        l = self.listenlow = Spinner(parent)
        l.min_max = minim, maxim
        l.show()
        l.value = low
        self.pack_end(l)

        h = self.listenhigh = Spinner(parent)
        h.min_max = minim, maxim
        h.show()
        h.value = high
        self.pack_end(h)

class Limits(Frame):
    def __init__(self, parent, session):
        Frame.__init__(self, parent)
        self.text = "Limits"
        self.collapse = True
        self.autocollapse = True

        self.size_hint_align = -1.0, 0.0

        b = Box(parent)
        ul = LimitWidget(self, "Upload limit",
            session.upload_rate_limit,
            session.set_upload_rate_limit)
        dl = LimitWidget(self, "Download limit",
            session.download_rate_limit,
            session.set_download_rate_limit)
        lul = LimitWidget(self, "Upload limit for local connections",
            session.local_upload_rate_limit,
            session.set_local_upload_rate_limit)
        ldl = LimitWidget(self, "Download limit for local connections",
            session.local_download_rate_limit,
            session.set_local_download_rate_limit)

        for w in ul, dl, lul, ldl:
            w.show()
            b.pack_end(w)

        self.content = b

class LimitWidget(Box):
    base = 1024
    units = ( "bytes/s", "KiB/s", "MiB/s", "GiB/s" )

    def __init__(self, parent, title, rfunc, wfunc):
        Box.__init__(self, parent)
        self.title = title

        self.horizontal = True
        self.size_hint_weight = 1.0, 0.0
        self.size_hint_align = -1.0, 0.0

        l = Label(parent)
        l.text = title
        l.size_hint_align = -1.0, 0.0
        l.show()

        self.save_timer = None

        usw = self.usw = UnitSpinner(parent, self.base, self.units)
        usw.size_hint_weight = 1.0, 0.0
        usw.size_hint_align = 1.0, 0.0
        usw.set_value(rfunc())
        usw.spinner.callback_changed_add(self.changed_cb, wfunc)
        usw.hoversel.callback_selected_add(self.changed_cb, wfunc)
        usw.show()

        self.pack_end(l)
        self.pack_end(usw)

    def changed_cb(self, widget, *args):
        wfunc = args[1] if len(args) == 2 else args[0]

        if self.save_timer is not None:
            self.save_timer.delete()

        self.save_timer = Timer(2.0, self.save_cb, wfunc)

    def save_cb(self, wfunc):
        v = int(self.usw.get_value())
        log.debug("Saving value {} for {}.".format(v, self.title))
        wfunc(v)
        return False

# TODO:
# max uploads?, max conns?, max half open conns?

# ip filter
